#ifndef BLD_BOOLOP_H
#define BLD_BOOLOP_H

/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <bld/op.h>

/***********************************************************************
class BLD_UnaryOperation

Abstract base class for unary operations.
***********************************************************************/
class BLD_UnaryOperation : public BLD_Operation
{
	public:
		BLD_UnaryOperation(BLD_Operation *o) : op(o) {}
		virtual ~BLD_UnaryOperation(void) {}

		virtual BLD_Value Perform(void) const { return BLD_Value(); }
		virtual GPL_String ToString(void) const { return GPL_String(); }

	protected:
		GPL_Pointer<BLD_Operation> op;
};

/***********************************************************************
class BLD_BinaryOperation

Abstract base class for unary operations.
***********************************************************************/
class BLD_BinaryOperation : public BLD_Operation
{
	public:
		BLD_BinaryOperation(BLD_Operation *l, BLD_Operation *r)
			: left(l), right(r) {}
		virtual ~BLD_BinaryOperation(void) {}

		virtual BLD_Value Perform(void) const { return BLD_Value(); }
		virtual GPL_String ToString(void) const { return GPL_String(); }

	protected:
		GPL_Pointer<BLD_Operation> left;
		GPL_Pointer<BLD_Operation> right;
};

/***********************************************************************
class BLD_NotOperation

Logically negates the outcome of the operation.
***********************************************************************/
class BLD_NotOperation : public BLD_UnaryOperation
{
	public:
		BLD_NotOperation(BLD_Operation *o) : BLD_UnaryOperation(o) {}
		virtual ~BLD_NotOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;
};

/***********************************************************************
class BLD_AndOperation

Logical intersection.
***********************************************************************/
class BLD_AndOperation : public BLD_BinaryOperation
{
	public:
		BLD_AndOperation(BLD_Operation *l, BLD_Operation *r)
			: BLD_BinaryOperation(l, r) {}
		virtual ~BLD_AndOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;
};

/***********************************************************************
class BLD_OrOperation

Logical union.
***********************************************************************/
class BLD_OrOperation : public BLD_BinaryOperation
{
	public:
		BLD_OrOperation(BLD_Operation *l, BLD_Operation *r)
			: BLD_BinaryOperation(l, r) {}
		virtual ~BLD_OrOperation(void) {}

		virtual BLD_Value Perform(void) const;
		virtual GPL_String ToString(void) const;
};

#endif /* BLD_BOOLOP_H */
