#ifndef IWS_ALIGN_H
#define IWS_ALIGN_H

/******************************************************************************

Coral Tree iws/align.h
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <gpl/gpl.h>

/***********************************************************************
class IWS_AnnotationAlignable
***********************************************************************/
class IWS_AnnotationAlignable
{
	protected:
		long textWidth;
		long textHeight;
		long annotationWidth;
		long annotationHeight;

	public:
		IWS_AnnotationAlignable()
		{
			textWidth = -1;
			textHeight = -1;
			annotationWidth = -1;
			annotationHeight = -1;
		}
virtual	~IWS_AnnotationAlignable(){}

virtual	void GetAnnotationSize(long & width, long & height) = 0;
virtual	void SetAnnotationSize(const long w, const long h) = 0;
virtual	const char *GetText() = 0;
};

/***********************************************************************
class IWS_AnnotationAlignableGroup
***********************************************************************/
class IWS_AnnotationAlignableGroup : public GPL_DoubleList<IWS_AnnotationAlignable>
{
	public:
		IWS_AnnotationAlignableGroup(){}
virtual	~IWS_AnnotationAlignableGroup(){}
		
		void Align();
};

#define IWS_AnnotationAlignable_DECLARATION()							\
		void UpdateSize();												\
		long GetAnnotationMode() const;									\
		char *GetAnnotationString(void) const;							\
		void GetAnnotationSize(long & width, long & height);			\
		void GetMaximumAnnotationSize(long & width, long & height);		\
		void GetMaximumSize(long & width, long & height);				\
		void SetAnnotationMode(const long mode);						\
		void SetAnnotation(const char *atitle);							\
		void SetTextSize(const long w, const long h);					\
		void SetAnnotationSize(const long w, const long h)

#define IWS_AnnotationAlignable_IMPLEMENTATION(className, superName)	\
long className::GetAnnotationMode(void) const							\
{																		\
	return ((className *) this)->GetLocalLook()->GetElement(			\
		WDS_STATE_ALL, WDS_ELEMENT_ANNOTATION_MODE);					\
}																		\
																		\
char *className::GetAnnotationString(void) const						\
{																		\
	return ((className *) this)->GetAnnotation()->Buffer();				\
}																		\
																		\
void className::GetAnnotationSize(long & width, long & height)			\
{																		\
	width = GetLocalLook()->GetElement(WDS_STATE_ALL, WDS_ELEMENT_ANNOTATION_X);	\
	height = GetLocalLook()->GetElement(WDS_STATE_ALL, WDS_ELEMENT_ANNOTATION_Y);	\
}																		\
																		\
void className::GetMaximumAnnotationSize(long & width, long & height)	\
{																		\
	long w = 0;															\
	long h = 0;															\
	GetAnnotationSize(w, h);											\
	if (w > width)														\
	{																	\
		width = w;														\
	}																	\
	if (h > height)														\
	{																	\
		height = h;														\
	}																	\
}																		\
																		\
void className::GetMaximumSize(long & width, long & height)				\
{																		\
	long w = 0;															\
	long h = 0;															\
	GetPrefSize(&w, &h);												\
	if (w > width)														\
	{																	\
		width = w;														\
	}																	\
	if (h > height)														\
	{																	\
		height = h;														\
	}																	\
}																		\
																		\
void className::SetAnnotationMode(const long mode)						\
{																		\
	GetLocalLook()->SetElement(WDS_STATE_ALL, WDS_ELEMENT_ANNOTATION_MODE, mode);	\
}																		\
																		\
void className::SetAnnotation(const char *atitle)						\
{																		\
	superName::SetAnnotation((char *) atitle);							\
	UpdateSize();														\
}																		\
																		\
void className::SetTextSize(const long w, const long h)					\
{																		\
	textWidth = w;														\
	textHeight = h;														\
	if (-1 != textWidth || -1 != textHeight)							\
	{																	\
		UpdateSize();													\
	}																	\
}																		\
																		\
void className::SetAnnotationSize(const long w, const long h)			\
{																		\
	annotationWidth = w;												\
	annotationHeight = h;												\
	if (-1 != annotationWidth || -1 != annotationHeight)				\
	{																	\
		UpdateSize();													\
	}																	\
}																		\
																		\
void className::UpdateSize(void)										\
{																		\
	long aw = 0;														\
	long ah = 0;														\
	register char *annotationString = GetAnnotationString();			\
	if ((char *) GPL_BAD_ADDRESS != GetAnnotationString())				\
	{																	\
		IWS_APP->GetApproximateSize(annotationString, aw, ah);			\
	}																	\
	if (-1 != annotationWidth)											\
	{																	\
		aw = annotationWidth;											\
	}																	\
	if (-1 != annotationHeight)											\
	{																	\
		ah = annotationHeight;											\
	}																	\
																		\
	long tw = 0;														\
	long th = 0;														\
	register const char *textString = GetText();						\
	if ((const char *) GPL_BAD_ADDRESS != textString)					\
	{																	\
		IWS_APP->GetApproximateSize(textString, tw, th);				\
	}																	\
	tw = textWidth;														\
	if (-1 != textHeight)												\
	{																	\
		th = textHeight;												\
	}																	\
																		\
	if (WDS_WIDGET_ANNOTATION_LEFT == GetAnnotationMode() ||			\
		WDS_WIDGET_ANNOTATION_RIGHT == GetAnnotationMode())				\
	{																	\
		if(-1!=tw) tw += aw;											\
		th = (th < ah) ? ah : th;										\
	}																	\
	else																\
	{																	\
		tw = (tw < aw) ? aw : tw;										\
		th += ah;														\
	}																	\
																		\
	SetLocalAnnotationSize(aw, ah);										\
	SetPrefSize(tw, th);												\
}

#endif /* IWS_ALIGN_H */
