/******************************************************************************

COPYRIGHT_HEADER

******************************************************************************/

#include <iws/iws.h>

IWS_Bag::IWS_Bag(void)
{
	WDS_Activation actv;
	actv.SetTargetWidget(this);
	SetStateChangeCallback(WDS_STATE_ALL,&IWS_Bag::TextListInducedCallback,&actv);
	WDS_TableEntryList *entrylist = new WDS_TableEntryList();
	SetTableEntryList(entrylist);
	SetEditable(WDS_TABLE_EDITABLE_ENTRY);
	SetHeaderLabels(1);
	needRebuild = FALSE;
}

IWS_Bag::~IWS_Bag(void)
{
	WDS_TableEntryList *entrylist = GetTableEntryList();
	if(entrylist) 
	{
		CleanUp();
		delete entrylist;
	}
}

GPL_String &IWS_Bag::GetEntry(long row, long col)
{
	WDS_TableEntryList *entrylist = GetTableEntryList();
	WDS_TableEntry *entry = entrylist->ToIndex(row);
	entryBuffer = entry->GetStringList()->ToIndex(col)->Buffer();
	return entryBuffer;
}

void IWS_Bag::UpdateValue(long row, long col)
{
	GPL_DoubleList<GPL_Tuple> tuplelist;
	Get(tuplelist);
	GPL_Tuple *tuple = tuplelist.ToIndex(row);
	if(tuple->GetValueList().GetCount() <= col)
	{
		return;
	}
	GPL_Value value = tuple->GetValueList()[col];
	value.SetPerType(GetEntry(row,col));
	tuple->GetValueList()[col] = value;
}


void IWS_Bag::TextListInducedCallback(WDS_Activation *actv)
{
	IWS_Bag *self = (IWS_Bag *)(actv->GetTargetWidget());
	if(actv->GetState() == WDS_STATE_0)
	{
		long row, col;
		self->GetLastEditIndices(&row, &col);
		self->UpdateValue(row, col);
		self->needRebuild = TRUE;
	}
}

void IWS_Bag::CleanUp()
{
	WDS_TableEntryList *entrylist = GetTableEntryList();
	WDS_TableEntry *entry;
	entrylist->ToHead();
	while(entry = entrylist->PostIncrement())
	{
		WDS_StringList *slist = entry->GetStringList();
		if(slist)
		{
			slist->ToHead();
			WDS_String *s;
			while(s = slist->PostIncrement())
			{
				slist->RemoveNode(s);
				delete s;
			}
			delete slist;
		}
		entrylist->RemoveNode(entry);
		delete entry;
	}
}

void IWS_Bag::Rebuild()
{
	CleanUp();
	WDS_TableEntryList *entrylist = GetTableEntryList();
	entrylist->Clear();
	GPL_DoubleList<GPL_Tuple> tuplelist;
	Get(tuplelist);
	tuplelist.ToHead();
	GPL_Tuple *tuple;
	while(tuple = tuplelist.PostIncrement())
	{
		WDS_TableEntry *entry = new WDS_TableEntry();
		WDS_StringList *slist = new WDS_StringList();
		entry->SetStringList(slist);
		GPL_List<GPL_Value> vlist = tuple->GetValueList();
		long i;
		for(i=0;i<vlist.GetCount();i++)
		{
			WDS_String *s = new WDS_String;
			s->NewBuffer((const char *)vlist[i].GetString());
			slist->AppendNode(s);
		}
		entrylist->AppendNode(entry);
	}
	RebuildColumns();
}

long IWS_Bag::PostOperate(EW_Event *e)
{
	if(needRebuild)
	{
		Rebuild();
		needRebuild = FALSE;
	}
	return 0;
}
