#ifndef IWS_FILEWIN_H
#define IWS_FILEWIN_H

/******************************************************************************

Coral Tree iws/filewin.h
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <iws/window.h>
#include <iws/getstr.h>
#include <iws/btnbar.h>

class IWS_FileSelector;

/***********************************************************************
class IWS_FileApplyHandler
***********************************************************************/
class IWS_FileApplyHandler : public IWS_Handler
{
	public:
		IWS_FileApplyHandler(void)
			: fileSelector((IWS_FileSelector *) GPL_BAD_ADDRESS) {}
		virtual ~IWS_FileApplyHandler(void) {}

		void SetFileSelector(IWS_FileSelector *f) { fileSelector = f; }
		IWS_FileSelector *GetFileSelector(void) { return fileSelector; }

		virtual void HandleEvent(const GPL_String &string)
		{
			wbaLog("IWS WARNING", "[%s] file apply not handled", 
					(const char *)string);
		}

	protected:
		IWS_FileSelector *fileSelector;
};

class IWS_FileSelectorWindow;

/***********************************************************************
class IWS_FileApplyRelayHandler
***********************************************************************/
class IWS_FileApplyRelayHandler : public IWS_FileApplyHandler
{
	public:
		IWS_FileApplyRelayHandler(IWS_FileSelectorWindow *w)
			: applyHandler((IWS_FileApplyHandler *) GPL_BAD_ADDRESS),
			window(w) {}
		virtual ~IWS_FileApplyRelayHandler(void) {}
		void SetHandler(IWS_FileApplyHandler *h)
		{
			applyHandler = h;
		}
		virtual void HandleEvent(const GPL_String & string);

	protected:
		IWS_FileApplyHandler *applyHandler;
		IWS_FileSelectorWindow *window;
};

/***********************************************************************
class IWS_FileSelectHandler
***********************************************************************/
class IWS_FileSelectHandler : public IWS_Handler
{
	public:
		IWS_FileSelectHandler(IWS_FileSelector *f) : fileSelector(f) {}
		virtual ~IWS_FileSelectHandler(void) {}

		virtual void HandleEvent(const GPL_String & string);

	protected:
		IWS_FileSelector *fileSelector;
};

/***********************************************************************
class IWS_FileUpdateDirectoryHandler
***********************************************************************/
class IWS_FileUpdateDirectoryHandler : public IWS_Handler
{
	public:
		IWS_FileUpdateDirectoryHandler(IWS_FileSelector *f) : fileSelector(f) {}
		virtual ~IWS_FileUpdateDirectoryHandler(void) {}

		virtual void HandleEvent(const GPL_String & string);

	protected:
		IWS_FileSelector *fileSelector;
};

/***********************************************************************
class IWS_FileUpdateFileHandler
***********************************************************************/
class IWS_FileUpdateFileHandler : public IWS_Handler
{
	public:
		IWS_FileUpdateFileHandler(IWS_FileSelector *f) : fileSelector(f) {}
		virtual ~IWS_FileUpdateFileHandler(void) {}

		virtual void HandleEvent(const GPL_String & string);

	protected:
		IWS_FileSelector *fileSelector;
};

/***********************************************************************
class IWS_FileUpdateFilterHandler
***********************************************************************/
class IWS_FileUpdateFilterHandler : public IWS_Handler
{
	public:
		IWS_FileUpdateFilterHandler(IWS_FileSelector *f) : fileSelector(f) {}
		virtual ~IWS_FileUpdateFilterHandler(void) {}

		virtual void HandleEvent(const GPL_String & string);

	protected:
		IWS_FileSelector *fileSelector;
};

/***********************************************************************
class IWS_FileSelector
***********************************************************************/
class IWS_FileSelector : public WDS_Partition, public GPL_FileSelector
{
	public:
		IWS_FileSelector(void);
		virtual ~IWS_FileSelector(void);

		virtual void SetDirectory(const GPL_String & d);
		virtual void SetFilterString(const GPL_String & f);
		virtual void SetFile(const GPL_String & f);

		virtual const GPL_String & GetSelectedFile(void);
		virtual void SetFilter(GPL_FileFilter *f);

		IWS_StringList *GetFileList(void) { return fileList; }
		IWS_GetString *GetDirField(void) { return dirField; }
		IWS_GetString *GetFileField(void) { return fileField; }
		IWS_GetString *GetFilterField(void) { return filterField; }

		void SetHandler(IWS_FileApplyHandler *handler)
		{
			applyHandler = handler;
		}

		void UpdateFileList(void);

		void Apply(WDS_Activation *actv, const GPL_String & string);

	protected:
		WDS_Form *verticalForm;
		IWS_GetString *dirField;
		IWS_GetString *fileField;
		IWS_GetString *filterField;
		IWS_StringList *fileList;

		IWS_FileSelectHandler *fileSelectHandler;
		IWS_FileUpdateDirectoryHandler *fileUpdateDirectoryHandler;
		IWS_FileUpdateFileHandler *fileUpdateFileHandler;
		IWS_FileUpdateFilterHandler *fileUpdateFilterHandler;
		IWS_FileApplyHandler *applyHandler;

		GPL_GlobRegExpFileFilter fileFilter;
		GPL_String selectedFile;

		void AlignAnnotations(void);
};

/***********************************************************************
class IWS_FileSelectorWindowButtonHandler
***********************************************************************/
class IWS_FileSelectorWindowButtonHandler : public IWS_Handler
{
	public:
		IWS_FileSelectorWindowButtonHandler(IWS_FileSelectorWindow *w)
			: window(w) {}
		virtual ~IWS_FileSelectorWindowButtonHandler(void) {}
	
		void HandleEvent(const GPL_String & string);

	protected:
		IWS_FileSelectorWindow *window;
};

/***********************************************************************
class IWS_FileSelectorWindow
***********************************************************************/
class IWS_FileSelectorWindow : public IWS_VerticalWindow
{
	public:
		IWS_FileSelectorWindow(void);
		virtual ~IWS_FileSelectorWindow(void);
		void SetApplyHandler(IWS_FileApplyHandler *h);

		IWS_FileSelector *GetFileSelector(void) { return fileSelector; }
		const GPL_String & GetSelectedFile(void)
		{
			return fileSelector->GetSelectedFile();
		}

	protected:
		IWS_FileSelector *fileSelector;
		IWS_ButtonRow *buttonRow;
		IWS_FileSelectorWindowButtonHandler *handler;
		IWS_FileApplyRelayHandler *relay;
};

#endif /* IWS_FILEWIN_H */
