
/******************************************************************************

COPYRIGHT_HEADER

******************************************************************************/

#include <iws/iws.h>

IWS_DefaultMenuHandler IWS_MenuBase::defaultHandler;

IWS_MenuBase::IWS_MenuBase()
{
}

IWS_MenuBase::~IWS_MenuBase()
{
}

GPL_List<GPL_String> IWS_MenuBase::GetMenuNodeNames()
{
	GPL_List<GPL_String> names;
	GPL_List<GPL_String> subNames;
	GPL_String name;
	EW_WidgetList * widgets = GetMenuNodeList ();
	widgets->ToHead();
	EW_Widget * widget;
	while((EW_Widget *)GPL_BAD_ADDRESS!=(widget = widgets->PostIncrement ()))
	{
		subNames = ((IWS_MenuNode *) widget)->GetMenuNodeNames ();
		if(0 == subNames.GetCount ())
		{
			names.Append (widget->GetName ());
			continue;
		}

		for(unsigned long i = 0; subNames.GetCount () > i; ++i)
		{
			name = widget->GetName ();
			name.Join ("->", subNames[i]);
			names.Append (name);
		}
	}

	return names;
}

IWS_MenuNode *IWS_MenuBase::FindMenuNode (const GPL_String & n) 
{
	GPL_String first;
	GPL_String second;
	n.Split ("->", first, second);

	if (0 < second.GetLength ()) 
	{
		IWS_MenuNode *node=(IWS_MenuNode *)FindWidget(GetMenuNodeList(), first);
		if ((IWS_MenuNode *) GPL_BAD_ADDRESS == node) 
		{

			return (IWS_MenuNode *) GPL_BAD_ADDRESS;
		}

		return node->FindMenuNode (second);
	}

	IWS_MenuNode *node =(IWS_MenuNode *)FindWidget(GetMenuNodeList(), first);

	return node;
}


/******************************************************************************
WARNING:  These two AddMenuNode()s are similar but not identical.  Be sure to 
duplicate significant changes in both.
******************************************************************************/
IWS_MenuNode *IWS_MenuBase::AddMenuNode(const GPL_String & n, IWS_MenuHandler *handler, WDS_Activation *actv)
{
	GPL_String first;
	GPL_String second;
	n.Split ("->", first, second);
	if (0 < second.GetLength ())
	{
		IWS_MenuNode *node=(IWS_MenuNode *)FindWidget(GetMenuNodeList(),first);
		if ((IWS_MenuNode *) GPL_BAD_ADDRESS == node)
		{
			node = new IWS_MenuNode ();
			node->SetName (first.GetData ());
			node->SetLabel (first.GetData ());
			AddMenuNodeProtected (node);
		}
		return node->AddMenuNode (second, handler);
	}

	if ((IWS_MenuNode *) GPL_BAD_ADDRESS != FindMenuNode (n))
	{
		return NULL;
	}

	IWS_MenuNode * node = new IWS_MenuNode ();
	node->SetName (n.GetData ());
	node->SetLabel (n.GetData ());
	if (handler)
	{
		node->SetHandler(handler, actv);
	}
	AddMenuNodeProtected (node);

	return node;
}

IWS_MenuNode *IWS_MenuBase::AddMenuNode(const GPL_String & n,void (*c) (WDS_Activation *),WDS_Activation * a)
{
	GPL_String first;
	GPL_String second;
	n.Split ("->", first, second);
	if (0 < second.GetLength ())
	{
		IWS_MenuNode *node=(IWS_MenuNode *)FindWidget(GetMenuNodeList(),first);
		if ((IWS_MenuNode *) GPL_BAD_ADDRESS == node)
		{
			node = new IWS_MenuNode ();
			node->SetName (first.GetData ());
			node->SetLabel (first.GetData ());
			AddMenuNodeProtected (node);
		}
		return node->AddMenuNode (second, c, a);
	}

	if ((IWS_MenuNode *) GPL_BAD_ADDRESS != FindMenuNode (n))
	{
		return NULL;
	}

	IWS_MenuNode * node = new IWS_MenuNode ();
	node->SetName (n.GetData ());
	node->SetLabel (n.GetData ());
	if ((void (*) (WDS_Activation *)) GPL_BAD_ADDRESS != c)
	{
		node->SetStateChangeCallback (WDS_STATE_0, c, a);
	}
	else
	{
		node->SetHandler(&defaultHandler);
	}
	AddMenuNodeProtected (node);

	return node;
}

GPL_Boolean IWS_MenuBase::RemoveMenuNode (const GPL_String & n)
{
	GPL_String first;
	GPL_String second;
	n.Split ("->", first, second);

	if (0 < second.GetLength ()) 
	{
		IWS_MenuNode *node=(IWS_MenuNode *)FindWidget(GetMenuNodeList(),first);
		if ((IWS_MenuNode *) GPL_BAD_ADDRESS == node) 
		{
			return FALSE;
		}

		return node->RemoveMenuNode (second);
	}

	IWS_MenuNode *node = (IWS_MenuNode *) FindWidget (GetMenuNodeList (), n);
	if ((IWS_MenuNode *) GPL_BAD_ADDRESS == node) 
	{
		return FALSE;
	}
	RemoveMenuNodeProtected (node);

	return TRUE;
}

void IWS_MenuBase::SetMenuNodeWeight(const GPL_String &n, const GPL_Integer weight) 
{
	IWS_MenuNode * node = FindMenuNode (n);
	if ((IWS_MenuNode *) GPL_BAD_ADDRESS != node) 
	{
		node->SetWeight (weight);
		SortByWeight (GetMenuNodeList ());
	}
}

void IWS_MenuBase::SortMenuNodesByWeight (void) 
{
	SortByWeight (GetMenuNodeList ());
}

void IWS_MenuBase::DisableMenuNode (const GPL_String & n) 
{
	IWS_MenuNode * node = FindMenuNode (n);
	if ((IWS_MenuNode *) GPL_BAD_ADDRESS != node) 
	{
		node->Disable ();
	}
}

void IWS_MenuBase::EnableMenuNode (const GPL_String & n) 
{
	IWS_MenuNode * node = FindMenuNode (n);
	if ((IWS_MenuNode *) GPL_BAD_ADDRESS != node) 
	{
		node->Enable ();
	}
}

WDS_Widget *IWS_MenuBase::FindWidget(EW_WidgetList *widgets,const GPL_String &n)
{
	widgets->ToHead ();
	WDS_Widget * widget;
	while ((WDS_Widget *) GPL_BAD_ADDRESS != (widget = (WDS_Widget *) widgets->PostIncrement ())) 
	{
		if (n == widget->GetName ()) 
		{
			return widget;
		}
	}

	return (WDS_Widget *) GPL_BAD_ADDRESS;
}

void IWS_MenuBase::SortByWeight (EW_WidgetList * widgets) 
{
	// Get all the nodes in the list.
	GPL_List<EW_Widget *> nodes;
	widgets->ToHead ();
	EW_Widget * widget;
	while ((EW_Widget *) GPL_BAD_ADDRESS!=(widget = widgets->PostIncrement ())) 
	{
		nodes.Append (widget);
		widgets->RemoveNode (widget);
	}

	// For each node...
	for (unsigned long i = 0; nodes.GetCount () > i; ++i) 
	{
		// Sort the sub-nodes first.
		((IWS_MenuNode *) nodes[i])->SortMenuNodesByWeight ();

		// Look for a node with a higher weight.
		GPL_Boolean inserted = FALSE;
		widgets->ToHead ();
		while ((EW_Widget *)GPL_BAD_ADDRESS!=(widget = widgets->GetCurrent ()))
		{
			// If there is a heavier node...
			if (((IWS_MenuNode *) nodes[i])->GetWeight () <
				((IWS_MenuNode *) widget)->GetWeight ()) 
			{
				// Insert the node before it.
				widgets->InsertBefore (nodes[i]);
				inserted = TRUE;
				break;
			}
			widgets->PostIncrement();
		}

		// If not previously inserted...
		if (!inserted) 
		{
			// Append to the end of the list.
			widgets->AppendNode (nodes[i]);
		}
	}
}

IWS_MenuNode::IWS_MenuNode (void) : weight (0), enabled (FALSE) 
{
	Enable ();
//	SetFlagsIncluding(WDS_POPUP_BUTTONLESS_POP);
}

IWS_MenuNode::~IWS_MenuNode (void) 
{
}

// TODO: This should be replaced by WDS_MenuNode::GetMenuForm().
WDS_MenuForm *IWS_MenuNode::GetMenuForm(void) 
{
	// Find the scope.
	EW_WidgetList * widgets = GetBaseWindow ()->GetWidgetList ();
	WDS_Widget * widget;
	widgets->ToHead ();
	while ((widget = (WDS_Widget *) widgets->PostIncrement ())) 
	{
		if (WDS_Scope::GetTypeCodeStatic () == widget->GetTypeCode ()) 
		{
			break;
		}
	}

	// Find the pop-up base.
	widgets = widget->GetWidgetList ();
	widgets->ToHead ();
	while ((widget = (WDS_Widget *) widgets->PostIncrement ())) 
	{
		if (WDS_PopupBase::GetTypeCodeStatic () == widget->GetTypeCode ()) 
		{
			break;
		}
	}

	// Find the form.
	widgets = widget->GetWidgetList ();
	widgets->ToHead ();
	while ((widget = (WDS_Widget *) widgets->PostIncrement ())) 
	{
		if (WDS_MenuForm::GetTypeCodeStatic () == widget->GetTypeCode ()) 
		{
			return (WDS_MenuForm *) widget;
		}
	}

	return (WDS_MenuForm *) GPL_BAD_ADDRESS;
}

void IWS_MenuNode::Disable (void) {

	if (enabled) 
	{
		// This policy (greyed-out and dormant) matches the OSF/Motif
		// standard.  Implementation of another policy would replace
		// the following two lines.
		SetGreyOut (TRUE);
		SetDormant (TRUE);

		enabled = FALSE;
	}
}

void IWS_MenuNode::Enable (void) 
{
	if (!enabled) 
	{
		// This policy (greyed-out and dormant) matches the OSF/Motif
		// standard.  Implementation of another policy would replace
		// the following two lines.
		SetGreyOut (FALSE);
		SetDormant (FALSE);

		enabled = TRUE;
	}
}

EW_WidgetList *IWS_MenuNode::GetMenuNodeList (void) 
{
	return GetMenuForm ()->GetWidgetList ();
}

void IWS_MenuNode::AddMenuNodeProtected (IWS_MenuNode * n) 
{
	AddWidget (n);
	SortByWeight (GetMenuNodeList ());
	n->SetPopItem(GetPopItem());
}

void IWS_MenuNode::RemoveMenuNodeProtected (IWS_MenuNode * n) 
{
	RemoveWidget (n);
}

GPL_String &IWS_MenuNode::GetFullname()
{
	fullname.Clear();
	fullname.Append(GetName());

	WDS_MenuNode *n;
	n = this;
	while(n->GetMaster())
	{
		n = n->GetMaster();
		fullname.Prepend("->");
		fullname.Prepend(n->GetName());
	}

	return fullname;
}

void IWS_MenuNode::SetHandler(IWS_MenuHandler *handler, WDS_Activation *actv)
{
	if(!handler) return;
	WDS_Activation a;
	if(!actv) actv = &a;
	actv->SetObject((void *)handler);
	SetStateChangeCallback(WDS_STATE_0, handler->Callback, actv);
}

IWS_MenuBar::IWS_MenuBar (void) 
{
	SetMode (WDS_FORM_SINGLE_ROW);
	SetName("IWS_MenuBar");
}

IWS_MenuBar::~IWS_MenuBar (void) 
{
}

EW_WidgetList *
IWS_MenuBar::GetMenuNodeList (void) 
{
	return GetWidgetList ();
}

void
IWS_MenuBar::AddMenuNodeProtected (IWS_MenuNode * n) 
{
	AddWidget (n);
	SortByWeight (GetMenuNodeList ());
}

void
IWS_MenuBar::RemoveMenuNodeProtected (IWS_MenuNode * n) 
{
	RemoveWidget (n);
}

GPL_String &IWS_MenuHandler::GetFullname()
{
	IWS_MenuNode *n;
	n = (IWS_MenuNode *)(activation->GetThisWidget());
	return n->GetFullname();
}

void IWS_DefaultMenuHandler::HandleEvent(const GPL_String &string)
{
	wbaLog("IWS WARNING", "[ %s (%s) ] menu item not handled", 
		(const char *)string, (const char *)(GetFullname()));
}

void IWS_MenuShowWindowHandler::HandleEvent(const GPL_String &string)
{
	IWS_MenuShowWindowHandler *self = (IWS_MenuShowWindowHandler *)
										GetActivation()->GetObject();
	if ((IWS_Window *) GPL_BAD_ADDRESS != self->window)
	{
		self->window->Show();
	}
}

void IWS_MenuHideWindowHandler::HandleEvent(const GPL_String &string)
{
	IWS_MenuHideWindowHandler *self = (IWS_MenuHideWindowHandler *)
										GetActivation()->GetObject();
	if ((IWS_Window *) GPL_BAD_ADDRESS != self->window)
	{
		self->window->Hide();
	}
}

void IWS_MenuExitApplicationHandler::HandleEvent(const GPL_String &string)
{
	IWS_APP->GetEventContext()->EndLoop();
}
