
/******************************************************************************

COPYRIGHT_HEADER

******************************************************************************/

class IWS_MenuNode;

class IWS_MenuHandler : public IWS_Handler
{
	public:
		IWS_MenuHandler(){}
virtual	~IWS_MenuHandler(){}

		GPL_String &GetFullname();
};

class IWS_DefaultMenuHandler : public IWS_MenuHandler
{
	public:
		IWS_DefaultMenuHandler(){}
virtual ~IWS_DefaultMenuHandler(){}

virtual	void HandleEvent(const GPL_String &string);
};

class IWS_MenuShowWindowHandler : public IWS_MenuHandler
{
	public:
		IWS_MenuShowWindowHandler(IWS_Window *w = (IWS_Window *) GPL_BAD_ADDRESS)
			: window(w) {}
		virtual ~IWS_MenuShowWindowHandler(void) {}

		virtual void HandleEvent(const GPL_String &string);

	protected:
		IWS_Window *window;
};

class IWS_MenuHideWindowHandler : public IWS_MenuHandler
{
	public:
		IWS_MenuHideWindowHandler(IWS_Window *w = (IWS_Window *) GPL_BAD_ADDRESS)
			: window(w) {}
		virtual ~IWS_MenuHideWindowHandler(void) {}

		virtual void HandleEvent(const GPL_String &string);

	protected:
		IWS_Window *window;
};

class IWS_MenuExitApplicationHandler : public IWS_MenuHandler
{
	public:
		IWS_MenuExitApplicationHandler(void) {}
		virtual ~IWS_MenuExitApplicationHandler(void) {}

		virtual void HandleEvent(const GPL_String &string);
};

class IWS_MenuBase
{
	private:
		static IWS_DefaultMenuHandler defaultHandler;
	protected:
virtual WDS_Widget * FindWidget (EW_WidgetList * widgets, const GPL_String & n);
virtual void SortByWeight (EW_WidgetList * widgets);
virtual EW_WidgetList * GetMenuNodeList (void) = 0;
virtual void AddMenuNodeProtected (IWS_MenuNode * n) = 0;
virtual void RemoveMenuNodeProtected (IWS_MenuNode * n) = 0;
	public:
		IWS_MenuBase (void);
virtual ~IWS_MenuBase (void);

virtual GPL_List<GPL_String> GetMenuNodeNames (void);

virtual IWS_MenuNode *FindMenuNode (const GPL_String & n);

virtual IWS_MenuNode *AddMenuNode (const GPL_String & n,
									IWS_MenuHandler *handler, 
									WDS_Activation *actv = NULL);
virtual IWS_MenuNode *AddMenuNode (const GPL_String & n,
									void (*c) (WDS_Activation *) =
										(void (*) (WDS_Activation *)) 
										GPL_BAD_ADDRESS,
									WDS_Activation * a =
										(WDS_Activation *) GPL_BAD_ADDRESS);
virtual GPL_Boolean RemoveMenuNode (const GPL_String & n);
virtual void SetMenuNodeWeight (const GPL_String & n,
									const GPL_Integer weight);
virtual void SortMenuNodesByWeight (void);
virtual void DisableMenuNode (const GPL_String & n);
virtual void EnableMenuNode (const GPL_String & n);
};

class IWS_MenuNode : public WDS_MenuNode, public IWS_MenuBase
{
	protected:
		GPL_Integer weight;
		GPL_Boolean enabled;
		GPL_String	fullname;

virtual EW_WidgetList * GetMenuNodeList (void);
virtual void AddMenuNodeProtected (IWS_MenuNode * n);
virtual void RemoveMenuNodeProtected (IWS_MenuNode * n);
	public:
		IWS_MenuNode (void);
virtual ~IWS_MenuNode (void);

inline 	GPL_Integer GetWeight (void) const { return weight; }
WDS_MenuForm * GetMenuForm(void);

inline 	void SetWeight (const GPL_Integer w) { weight = w; }
		void Disable (void);
		void Enable (void);

		GPL_String &GetFullname();

		void SetHandler(IWS_MenuHandler *handler, WDS_Activation *actv = NULL);
};

class IWS_MenuBar : public WDS_MenuBar, public IWS_MenuBase
{
	protected:
virtual EW_WidgetList * GetMenuNodeList (void);
virtual void AddMenuNodeProtected (IWS_MenuNode * n);
virtual void RemoveMenuNodeProtected (IWS_MenuNode * n);
	public:
		IWS_MenuBar (void);
virtual ~IWS_MenuBar (void);
};


