/******************************************************************************

Coral Tree iws/pick.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <iws/pick.h>
#include <iws/app.h>

/***********************************************************************
class IWS_Pick
***********************************************************************/
IWS_DefaultPickHandler IWS_Pick::defaultHandler;

IWS_Pick::IWS_Pick(const GPL_String & n,const long tw,const long th,
					const long aw, const long ah)
{
	IWS_PickConstructor(tw, th, aw, ah);
	GPL_String s = n;
	SetName((char *) s);
	SetAnnotation((const char *) s);
}

IWS_Pick::IWS_Pick(const long tw,const long th, const long aw, const long ah)
{
	IWS_PickConstructor(tw, th, aw, ah);
}

IWS_Pick::~IWS_Pick(void)
{
}

void IWS_Pick::IWS_PickConstructor(const long tw, const long th,
									const long aw, const long ah)
{
	SetGeneralFlags(WDS_WIDGET_FLAGS_AUTOSIZE_Y);
	if (-1 == tw)
	{
		SetBounds(EW_HORIZONTAL, EW_BOUNDS_NONE);
	}
	SetAnnotationMode(WDS_WIDGET_ANNOTATION_LEFT);
	SetTextSize(tw, th);
	SetAnnotationSize(aw, ah);
	SetHandler(&defaultHandler);
}

void IWS_Pick::AppendEntry(const GPL_String &n)
{
	WDS_StringList *slist = GetPickList()->GetStringList();
	WDS_String *string = new WDS_String();
	string->NewBuffer((const char *) n);
	slist->AppendNode(string);
}

long IWS_Pick::RemoveEntry(const GPL_String &n)
{
	WDS_StringList *slist = GetPickList()->GetStringList();
	WDS_String *string;
	slist->ToHead();
	while (string = slist->PostIncrement())
	{
		GPL_String s = n;
		if (!string->Compare((char *) s))
		{
			slist->RemoveNode(string);
			delete string;
			return TRUE;
		}
	}
	return FALSE;
}



/******************************************************************************
void	IWS_Pick::ClearEntries(void)

******************************************************************************/
void IWS_Pick::ClearEntries(void)
	{
	WDS_StringList *slist = GetPickList()->GetStringList();
	slist->Clear();
	}



IWS_AnnotationAlignable_IMPLEMENTATION(IWS_Pick, WDS_Pick);

const GPL_String & IWS_Pick::GetSelected(void)
{
	selected = GetLabel();
	return selected;
}

void IWS_Pick::SetHandler(IWS_PickHandler *handler, WDS_Activation *actv)
{
	if ((IWS_Handler *) GPL_BAD_ADDRESS == handler)
	{
		return;
	}
	WDS_Activation a;
	if ((WDS_Activation *) GPL_BAD_ADDRESS == actv)
	{
		actv = &a;
	}
	actv->SetObject((void *) handler);
	SetStateChangeCallback(WDS_STATE_0, handler->Callback, actv);
}

const char *IWS_Pick::GetText(void)
{
	return (GetSelected().GetData());
}

/***********************************************************************
class IWS_PickHandler
***********************************************************************/
void IWS_PickHandler::SetString(GPL_String &string)
{
	string = GetActivation()->GetThisWidget()->GetLabel();
}
