/******************************************************************************

Coral Tree iws/window.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <iws/window.h>
#include <iws/app.h>

IWS_Window::IWS_Window(const char *title)
{
	/* Add window to main event context */
	IWS_Application::eventContext->AddWindow(this);
	rootScope = new WDS_Scope();
	rootScope->SetBounds(EW_BOUNDS_FILL_PARENT);
	AddWidget(rootScope);
	SetSize(480,320);
	if ((const char *) GPL_BAD_ADDRESS != title)
	{
		SetTitle((char *) title);
	}
}

IWS_Window::~IWS_Window(void)
{
	delete rootScope;
	if(!GetEventContext())
	{
		wbaLog("IWS WARNING", "IWS_Window left without context.  Possibly newed w/o delete");
	}
	else
	{
		IWS_Application::eventContext->RemoveWindow(this);
	}
}
	
void IWS_Window::IncludeWidget(WDS_Widget *widget)
{
	rootScope->AddWidget(widget);
}

GPL_Boolean IWS_Window::IsShown(void) const
{
	return IsWindowShown((EW_Window *) this);
}

void IWS_Window::Show(void)
{
	ShowWindow((EW_Window *) this);
}

void IWS_Window::Hide(void)
{
	HideWindow((EW_Window *) this);
}

GPL_Boolean IWS_Window::IsWindowShown(EW_Window * window)
{
	if(window->GetOpened() && window->GetMapped())
	{
		return TRUE;
	}
	return FALSE;
}

void IWS_Window::ShowWindow(EW_Window * window)
{
	// If the window is not opened...
	if (!window->GetOpened ()) {

		// Open the window without mapping it.
		// This avoids flicker.
		window->SetFlags (EW_WINDOW_NOMAP);
		window->Open ();
	}

	// If the window is not mapped..
	if (!window->GetMapped ()) {

		// Map the window.
		window->Map ();
	}
}

void IWS_Window::HideWindow(EW_Window * window)
{
	// If the window is mapped...
	if (window->GetMapped ()) {

		// Unmap the window.
		window->UnMap ();
	}
}

IWS_VerticalWindow::IWS_VerticalWindow(const char *title) : IWS_Window(title)
{
	rootPartition = new WDS_Partition();
	rootPartition->SetDirection(WDS_VERTICAL);
	rootPartition->SetBounds(EW_BOUNDS_FILL_PARENT);
	rootPartition->SetFlags(WDS_PARTITION_FLAGS_ATTACH);
	rootScope->AddWidget(rootPartition);
}

IWS_VerticalWindow::~IWS_VerticalWindow(void)
{
	delete rootPartition;
}

void IWS_VerticalWindow::IncludeWidget(WDS_Widget *widget)
{
	rootPartition->AddWidget(widget);
}

IWS_HorizontalWindow::IWS_HorizontalWindow(void)
{
	rootPartition = new WDS_Partition();
	rootPartition->SetDirection(WDS_HORIZONTAL);
	rootPartition->SetBounds(EW_BOUNDS_FILL_PARENT);
	rootPartition->SetFlags(WDS_PARTITION_FLAGS_ATTACH);
	rootScope->AddWidget(rootPartition);
}

IWS_HorizontalWindow::~IWS_HorizontalWindow(void)
{
	delete rootPartition;
}

void IWS_HorizontalWindow::IncludeWidget(WDS_Widget *widget)
{
	rootPartition->AddWidget(widget);
}

void IWS_ShowWindowHandler::HandleEvent(const GPL_String &string)
{
	IWS_ShowWindowHandler *self = (IWS_ShowWindowHandler *)
									GetActivation()->GetObject();
	if ((IWS_Window *) GPL_BAD_ADDRESS != self->window)
	{
		self->window->Show();
	}
}

void IWS_HideWindowHandler::HandleEvent(const GPL_String &string)
{
	IWS_HideWindowHandler *self = (IWS_HideWindowHandler *)
									GetActivation()->GetObject();
	if ((IWS_Window *) GPL_BAD_ADDRESS != self->window)
	{
		self->window->Hide();
	}
}
