
/******************************************************************************

COPYRIGHT_HEADER

******************************************************************************/

#include <iws/iws.h>

class IWS_ExampleApp : public IWS_Application
{
	private:
		IWS_VerticalWindow		*base_window;
		IWS_FileSelectorWindow	*file_window;
	public:
		IWS_ExampleApp();
virtual	~IWS_ExampleApp();

virtual const char * GetName(void) const { return "IWS Test Program"; }
virtual const char * GetParameterUsage(void) const { return ""; }

virtual GPL_Boolean Initialize(void);
virtual void Finalize(void);

virtual void Process();

};

class FileHandler : public IWS_FileApplyHandler
{
	public:
		void HandleEvent(const GPL_String &string)
		{
			wbaLog("IWS LOG", "file [%s] is selected [%s]", (const char *)(GetFileSelector()->GetSelectedFile()), (const char *)string);
		}
};

IWS_ExampleApp::IWS_ExampleApp()
{
	base_window = NULL;
	file_window = NULL;
}

IWS_ExampleApp::~IWS_ExampleApp()
{
}

void dummyCallback(WDS_Activation * activation)
{
	wbaLog("IWS WARNING", "dummy callback");
}

class FileMenuHandler : public IWS_MenuHandler
{
	private:
		IWS_FileSelectorWindow 	filewin;
		FileHandler				fileHandler;
	public:
		FileMenuHandler()
		{
			filewin.SetApplyHandler(&fileHandler);
			filewin.SetTitle("File Selector Window");
		}
virtual	~FileMenuHandler()
		{
		}
		void HandleEvent(const GPL_String &string)
		{
			if(string == "Open")
			{
				filewin.Show();
			}
			else if(string == "Load")
			{
				filewin.Show();
			}
			else if(string == "Close")
			{
				IWS_APP->eventContext->EndLoop();
			}
		}
};

GPL_Boolean IWS_ExampleApp::Initialize()
{
	GPL_Boolean return_val;
	return_val = IWS_Application::Initialize();

	base_window = new IWS_VerticalWindow("The test GUI known as XIWS");
	base_window->SetGeometry(0, 0, 500, 750);
	base_window->Show();
fprintf(stderr,"A\n");

#if 0
	IWS_Button *button;
	button = new IWS_Button();
	button->SetLabel("button");
	button->SetSizeMode(IWS_SIZE_MODE_FILL, IWS_SIZE_MODE_WRAP);
	base_window->IncludeWidget(button);
#endif
fprintf(stderr,"1\n");

	IWS_Handler *handler;
	handler = new IWS_Handler();
fprintf(stderr,"2\n");

	FileMenuHandler *fmenuhandler;
	fmenuhandler = new FileMenuHandler();
fprintf(stderr,"3\n");

	IWS_MenuBar *menu;
	menu = new IWS_MenuBar();
	menu->AddMenuNode("File->Open", fmenuhandler);
	menu->AddMenuNode("File->Load", fmenuhandler);
	menu->AddMenuNode("File->Close", fmenuhandler);
	menu->AddMenuNode("Help->About->this");
	menu->AddMenuNode("Help->About->that");
	menu->AddMenuNode("Help->Index");
	menu->SetMenuNodeWeight("Help",10);
	menu->AddMenuNode("Options->General");
	WDS_Activation actv;
	actv.SetCode(1234);
	menu->AddMenuNode("Options->Specific", &dummyCallback, &actv);
	base_window->IncludeWidget(menu);
fprintf(stderr,"4\n");

	IWS_FileSelectorButton *fileButton = new IWS_FileSelectorButton();
	base_window->IncludeWidget(fileButton);
fprintf(stderr,"5\n");

	IWS_ButtonBar *bar;
	IWS_Button *btn;
	bar = new IWS_ButtonBar();
	btn = bar->AddButton("Kill"); btn->SetImage(WDS_ICON_circle);
	btn = bar->AddButton("him"); btn->SetImage(WDS_ICON_check);
	btn = bar->AddButton("my");btn->SetImage(WDS_ICON_flash0);
	btn = bar->AddButton("rembots");btn->SetImage(WDS_ICON_mail);
	base_window->IncludeWidget(bar);

#if 0
	IWS_Button *button;
	button = new IWS_Button("bigbutton");
	button->SetSizeMode(IWS_SIZE_MODE_FILL, IWS_SIZE_MODE_FILL);
	base_window->IncludeWidget(button);
	button->SetAccelerator(EW_KEY_F(1));

	IWS_GetString *getstring;
	getstring = new IWS_GetString("getstring");
	base_window->IncludeWidget(getstring);

	IWS_StringList *stringlist;
	stringlist = new IWS_StringList("stringlist");
	base_window->IncludeWidget(stringlist);
	stringlist->AddText("I");
	stringlist->AddText("thought");
	stringlist->AddText("I saw");
	stringlist->AddText("a puttytat");
#endif
fprintf(stderr,"B\n");
	
	IWS_ButtonRow *bar2;
	bar2 = new IWS_ButtonRow();
	bar2->AddButton("I", handler);
	bar2->AddButton("did", handler);
	bar2->AddButton(" I ", handler);
	bar2->AddButton(" did ", handler);
	base_window->IncludeWidget(bar2);
	IWS_LookButton *lookbutton;
	//bar2->AddWidget(&lookbutton);
	lookbutton = new IWS_LookButton();
	base_window->IncludeWidget(lookbutton);

	IWS_MenuPopup *popup;
	popup = new IWS_MenuPopup(bar2);
	popup->SetName("popup");
	popup->AddMenuNode("over");
	popup->AddMenuNode("out->doors");
	popup->AddMenuNode("under->wear");

#if 0
	WDS_Button *butn;
	butn = bar2->FindButton("did");
	IWS_MenuPopup *popup3;
	popup3 = new IWS_MenuPopup(butn);
	popup3->SetName("popup3");
	popup3->AddMenuNode("red");
	popup3->AddMenuNode("white");
	popup3->AddMenuNode("blue->water");
#endif

#if 0
	IWS_ButtonRow *bar3;
	bar3 = new IWS_ButtonRow();
	bar3->AddButton("ok", handler);
	bar3->AddButton("apply", handler);
	bar3->AddButton("close", handler);
	base_window->IncludeWidget(bar3);
#endif

	IWS_Pick *pick;
	pick = new IWS_Pick("pick_list:");
	pick->SetLabel("unset");
	pick->AppendEntry("first");
	pick->AppendEntry("second");
	pick->AppendEntry("third");
	pick->AppendEntry("forth");
	base_window->IncludeWidget(pick);

#if 0
	IWS_RadioRow *rbar;
	rbar = new IWS_RadioRow();
	rbar->AddRadio("one");
	rbar->AddRadio("two");
	rbar->AddToggle("three");
	rbar->AddRadio("four");
	rbar->AddRadio("five");
	rbar->AddRadio("six");
	rbar->AddRadio("seven");
	rbar->AddRadio("eight");
	base_window->IncludeWidget(rbar);
#endif

#if 0
	IWS_FileSelector *fselect;
	fselect = new IWS_FileSelector();
	FileHandler *fhandler;
	fhandler = new FileHandler();
	fselect->SetHandler(fhandler);
	base_window->IncludeWidget(fselect);

	IWS_MenuPopup *popup2;
	popup2 = new IWS_MenuPopup(fselect);
	popup2->SetName("popup2");
	popup2->AddMenuNode("red");
	popup2->AddMenuNode("rat->like");
	popup2->AddMenuNode("cat");
#endif

fprintf(stderr,"C\n");
	IWS_Bag *bag;
	bag = new IWS_Bag();
	GPL_Tuple t;
	t.Set("SSSSS", "AAAAAAA","BBBBBBB","CCCCCCC","DDDDDDD","EEEEEEE");
	bag->Add(t);
	t.Set("IISIR", 1, 2, "hairy", 3, 4.5, .01);
	bag->Add(t);
	t.Set("III", 4, 5, 6);
	bag->Add(t);
	t.Set("SIF", "hairy", 3, 4.5, .01);
	bag->Add(t);
	bag->Rebuild();
	base_window->IncludeWidget(bag);

	IWS_TipBar *tipbar;
	tipbar = new IWS_TipBar(base_window);
	base_window->IncludeWidget(tipbar);

	IWS_StatusButton *statusButton;
	statusButton = new IWS_StatusButton();
	base_window->IncludeWidget(statusButton);
	statusButton->SetStatus("ready to roll");
	statusButton->AttachLogger();

	pick->SetTipMode(WDS_TIP_ON);
	pick->SetTipText("Pick A Number");

	tipbar->SetTipMode(WDS_TIP_ON);
	tipbar->SetTipText("No tip for tips");
fprintf(stderr,"D\n");

	return return_val;
}

void IWS_ExampleApp::Finalize()
{
	//delete base_window;
	delete base_window;

	IWS_Application::Finalize();
}

void IWS_ExampleApp::Process()
{
	wbaLog("IWS LOG", "base window shown %d", base_window->IsShown());
}


int main(const int argc, const char **argv, const char **envp)
{
#define IWS_USING_CLASS
#ifdef IWS_USING_CLASS
	IWS_ExampleApp app;

	app.LoadConfiguration(argc,argv,envp);

	return app.Main();
#else
	IWS_APP->LoadConfiguration(argc,argv,envp);
	IWS_VerticalWindow window;
	window.Show();
	return IWS_APP->Main();
#endif
}

