
/******************************************************************************

COPYRIGHT_HEADER

******************************************************************************/

#include <mte/mte.h>

MTE_CondVar::MTE_CondVar()
{
#if MTE_IMPL==MTE_SPROC
#elif MTE_IMPL==MTE_PTHREAD
#ifdef MTE_MUTEX_CV_IMPL
	int n;
	n = pthread_cond_init(&mutex, NULL);
	if(n)
	{
		wbaLog("MTE ERROR", "could not create pthread condition variable");
	}
#else
	signalMutex.Lock();
#endif
#elif MTE_IMPL==MTE_WIN32
	wbaLog("MTE ERROR", "MTE_CondVar not yet implemented in Win32");
#else
#endif
}

MTE_CondVar::MTE_CondVar(MTE_CondVar *c)
{
#if MTE_IMPL==MTE_SPROC
	wbaLog("MTE ERROR", "MTE_CondVar not yet implemented in IRIX sproc");
#elif MTE_IMPL==MTE_PTHREAD
#ifdef MTE_MUTEX_CV_IMPL
	cond = c->cond;
#endif
#elif MTE_IMPL==MTE_WIN32
#else
#endif
}

MTE_CondVar::~MTE_CondVar()
{
#if MTE_IMPL==MTE_SPROC
#elif MTE_IMPL==MTE_PTHREAD
#ifdef MTE_MUTEX_CV_IMPL
	if(pthread_cond_destroy(&cond))
	{
		wbaLog("MTE ERROR", "could not destroy pthread mutex");
	}
#endif
#elif MTE_IMPL==MTE_WIN32
#else
#endif
}

GPL_Boolean MTE_CondVar::Wait(MTE_Mutex *mutex)
{
#if MTE_IMPL==MTE_SPROC
	return TRUE;
#elif MTE_IMPL==MTE_PTHREAD
#ifdef MTE_MUTEX_CV_IMPL
	if(pthread_cond_wait(&cond,mutex->GetPthreadMutex()))
	{
		return FALSE;
	}
	return TRUE;
#else
	atomicMutex.Lock();
	mutex->Unlock();
	signalMutex.Lock();
	atomicMutex.Unlock();
	mutex->Lock();
	return TRUE;
#endif
#elif MTE_IMPL==MTE_WIN32
	return TRUE;
#else
	return TRUE;
#endif
}

GPL_Boolean MTE_CondVar::Signal()
{
#if MTE_IMPL==MTE_SPROC
	signalMutex.Unlock();
	atomicMutex.Lock(); // wait so that mutex.Lock() in Wait will happen
	atomicMutex.Unlock();
	return TRUE;
#elif MTE_IMPL==MTE_PTHREAD
#ifdef MTE_MUTEX_CV_IMPL
	if(pthread_cond_signal(&cond))
	{
		return FALSE;
	}
	return TRUE;
#else
	signalMutex.Unlock();
	return TRUE;
#endif
#elif MTE_IMPL==MTE_WIN32
	signalMutex.Unlock();
	atomicMutex.Lock(); // wait so that mutex.Lock() in Wait will happen
	atomicMutex.Unlock();
	return TRUE;
#else
	return TRUE;
#endif
}

