#ifndef NL_CHANNEL_H
#define NL_CHANNEL_H

/******************************************************************************

Coral Tree nl/channel.h
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <nl/system.h>
#include <nl/nl_proto.h>

#define EXIT(a,e)			NL_PRINT(NL_NL,NL_FATAL,a)
#define ERR_IF(n,a)			if(n)NL_PRINT(NL_NL,NL_ERROR,a)
#define	NL_MAX_HOSTNAME_LEN    32
#define NL_PRINT(source,type,string)	nlPrint(source,type,string,__FILE__,__LINE__)

/* NL_PRINT sources */
#define	NL_APP		0	/* application */
#define	NL_NL		1	/* Net Layer itself */
#define	NL_OS		2	/* Operating System */

/* NL_PRINT types */
#define	NL_LOG		0	/* verbose log updates */
#define	NL_WARN		1	/* possible problem */
#define	NL_ERROR	2	/* definite problem */
#define	NL_FATAL	3	/* fatal problem */

#define NL_CHANNEL_NOERROR		 0
#define NL_CHANNEL_ERROR		-1
#define NL_CHANNEL_WOULD_BLOCK	-2
#define NL_CHANNEL_TIMED_OUT	-3

/* connection type */
#define NL_CONNECTED		1
#define NL_CONNECTIONLESS	2

/* domains */
#define NL_INET				1

/* protocols */
#define NL_IP				1

/* inet_port[3] and inet_address[3] indexes */
#define NL_LOCAL			0
#define NL_REMOTE			1
#define NL_RETURN			2

/***********************************************************************
class NL_Channel

***********************************************************************/
class NL_Channel
{
	public:
		static long		count;

	private:
		uint16				inet_port[3];
		NL_HOST_INET_ADDR	inet_address[3];
		SOCKET_HANDLE_T		ip_handle;
		uint32				protocol;
		uint32 				domain;
		long				connection_type;
		long 				open;
		long				q_specified;
		long				valid;
		short 				blocking;
		long				last_errno;

		void	SetValid(long set)
				{
					// Set invalid on significant error.
					valid = set;
					if (!valid)
					{
						Close();
					}
				}
		uint16	InetGetPort(char *service_name);

	public:
		NL_Channel(void);
		virtual ~NL_Channel(void);

		long	GetValid(void) const { return valid; };
		SOCKET_HANDLE_T GetIPHandle(void) const { return ip_handle; }
		uint32	GetProtocol(void) const { return protocol; }
		uint32 	GetDomain(void) const { return domain; }
		char 	*GetHostname(long location) const;
		uint32	GetAddress(long location) const;
		long	GetPort(long location, char *service_name,
						long service_name_len) const;
		uint16	GetPort(long location) const;
		/* DO NOT USE GetLastError() due to NT name collision */
		long		CheckLastError(void) const { return last_errno; }
		long		IsOpen(void) const { return open; }

		uint16	Port(uint32 index) const
				{
					if (2 >= index)
					{
						return inet_port[index];
					}
					else
					{
						return 0;
					}
				}
		NL_HOST_INET_ADDR	Address(uint32 index) const
							{
								if (2 >= index)
								{
									return inet_address[index];
								}
								else
								{
									return 0;
								}
							}

		void	SetConnectionType(long setting) { connection_type = setting; } 
		long	SetBlocking(short set);
		long	SetAddress(long location, char *machine_name);
		long	SetAddress(long location, uint32 machine_number);
		long	SetPort(long location, char *service_name);
		long	SetPort(long location, uint16 port);
		void 	Set(uint32 a_domain, uint32 a_protocol, long type,
					uint16 local_port)
				{
					AddDomain(a_domain);
					AddProtocol(a_protocol);
					SetConnectionType(type);
					SetPort(NL_LOCAL, local_port);
				}
		void 	Set(uint32 a_domain, uint32 a_protocol, long type,
					uint16 local_port, char *remote_machine,
					uint16 remote_port)
				{
					AddDomain(a_domain);
					AddProtocol(a_protocol);
					SetConnectionType(type);
					SetPort(NL_LOCAL, local_port);
					SetAddress(NL_REMOTE, remote_machine);
					SetPort(NL_REMOTE, remote_port);
				}
		void 	Set(uint32 a_domain, uint32 a_protocol, long type,
					uint16 local_port, NL_HOST_INET_ADDR remote_address,
					uint16 remote_port)
				{
					AddDomain(a_domain);
					AddProtocol(a_protocol);
					SetConnectionType(type);
					SetPort(NL_LOCAL, local_port);
					SetAddress(NL_REMOTE, remote_address);
					SetPort(NL_REMOTE, remote_port);
				}

		uint16	FindLocalPort(void);
		void	AddProtocol(uint32 setting) { protocol = protocol | setting; }
		void	RemoveProtocol(uint32 setting) { protocol = protocol | ~setting; }
		void	AddDomain(uint32 setting) { domain = domain | setting; }
		void	RemoveDomain(uint32 setting) { domain = domain | ~setting; }

		// Actions.
		long		Open(void);
		void		Close(void);
		long		Send(char *buffer, long nbytes);
		long		Receive(char *buffer, long *nbytes, long msec_timeout);
		long		Receive(char *buffer, long nbytes, long msec_timeout);
		long 		Receive(char *buffer, long nbytes)
					{
						long tmp_nbytes;
						tmp_nbytes = nbytes;
						return Receive(buffer, &tmp_nbytes);
					}
		long		Receive(char *buffer, long *nbytes)
					{
						return Receive(buffer, nbytes, 0);
					}
		long		Connect(void);

		long		SpecifyQueue(void);
		NL_Channel	*Accept(void);
};

#endif /* NL_CHANNEL_H */
