/* A WIDGET SET */
/*
 * Right now this widget set's functionality is under development
 * so the graphics are just black and white lines
 */

/* FALSE == 0 */
/* TRUE == 1 */
#define TRUE_EXCLUDE	2

#define EW_BACKGROUND		WHITE
#define EW_FOREGROUND		BLACK
#define EW_LIGHTEDGE		130
#define EW_DARKEDGE			131
#define	EW_DARKGROUND		132
#define EW_LIGHTGROUND		133

#define CLX			9
#define CLY			14

/*****************************************************************************
EW_Button

A plain push button.
active == TRUE when pressed
active == FALSE when not pressed

void (*displayfunction)(long x,long y,long w,long h,void *displayarg)
	This function (if set) is called when the button is drawn.
*****************************************************************************/
class EW_Button : public EW_Widget
{
	private:
		EW_String label;

		void (*displayfunction)(EW_Window*, long, long, long, long, void *);
		void *displayarg;

		long q;

	public:
		EW_Button();
		long Operate(EW_Event *event,void *userdata);
		void Draw(EW_Redraw *redraw);
		void SetLabel(char *a_string);
		void SetDisplayFunction(void (*a_displayfunction)(EW_Window*,long,long,long,long,void *))
			{displayfunction = a_displayfunction;}
		void SetDisplayArg(void *a_displayarg){displayarg = a_displayarg;}
		long Resize(void *userdata);
};

/*****************************************************************************
EW_Toggle

A toggle button
active == TRUE when pressed
active == FALSE when not pressed
*****************************************************************************/
class EW_Toggle : public EW_Widget
{
	private:
		EW_String label;
	public:
		EW_Toggle();
		long Operate(EW_Event *event,void *userdata);
		void Draw(EW_Redraw *redraw);
		void SetLabel(char *a_string);
};


/*****************************************************************************
EW_GetString

A string entry widget
active == TRUE is an invalid state for this widget
active == FALSE when simply displaying string
active == TRUE_EXCLUDE when entering text

When a EW_GetString is activated active is set to TRUE_EXCLUDE.  All
sibling widgets with active == TRUE_EXCLUDE are deactivated.  Text is entered
while the mouse is within the parent widget.  Therefore, only one chlid
EW_GetString of a widget will get text entry at a time.

A side effect of this approach is that a EW_GetString that is the within
an ancestor widget of another EW_GetString can possibly accept input
simulataneously with the other widget.  
*****************************************************************************/
#define STRING_MODE	0
#define FLOAT_MODE 1
#define STRING_BUFFER 128
class EW_GetString : public EW_Widget
{
	private:
		EW_String		buffer;
		EW_String		string;

#if FALSE
		char	buffer[STRING_BUFFER];	// to hold new string
		char	string[STRING_BUFFER];	// to hold string
#endif

		long	mode; 					// STRING_MODE || FLOAT_MODE
		void	(*callback)(EW_Widget *, char *);	// the callback function
		long	c;						// character position
		long 	maxlength;
	public:
		EW_GetString();
		void Activate();
		void DeActivate();
		long Operate(EW_Event *event,void *userdata);
		void Draw(EW_Redraw *redraw);
		void RegisterCallback(void (*a_callback)(EW_Widget *, char *));
		void SetString(char *a_string);
};


/*****************************************************************************
EW_Blank

A widget that simply clears its area to some color.
The default color is EW_BACKGROUND.
*****************************************************************************/
class EW_Blank : public EW_Widget
{
	private:
		long color_index;
	public:
		EW_Blank();
		void Draw(EW_Redraw *redraw);
		long Resize(void *userdata);
		void SetColor(long a_color);
};

/*****************************************************************************
EW_Radio

This widget is design to contain EW_Toggle widgets.
During Operate() all child widgets are DeActivate()ed.
During PostOperate() an active child is searched for and its index in
widgetlist is set in active_index.

The effect when used with EW_Toggle widgets is that only one toggle can
be active at a time.

active_index == -1 when no child widgets are active.
IF more than one child widget is active, active_index will be the index of
the first one in widgetlist.

WARNING: The state of an EW_Radio is not accurate between Operate() and 
PostOperate() of the EW_Radio.  (Meaning all Operate()s and PostOperate()s
of descendents)
*****************************************************************************/
class EW_Radio : public EW_Widget
{
	protected:
		long active_index;
	public:
		EW_Radio();
		long Operate(EW_Event *event,void *userdata);
		long PostOperate(EW_Event *event,void *userdata);
		long ActiveIndex();
};


/*****************************************************************************
EW_NumberEntry

This widget contains 2 specialized button widgets for incrementing and
decrementing a value.  The value itself is a float, and is registered with
the widget with SetNumber(float *a_float).  This widget also contains
a EW_GetString widget which is where numbers are typed in.

This widget should be set up with the SetUp function.  SetGeometry and 
Attach should be avoided.

   ----------------------------------
   | label  | entry field | up | dn |
   ----------------------------------

*****************************************************************************/

class EW_IncrButton : public EW_Button
{
	public:
		EW_IncrButton(){SetDisplayFunction(ewDfUpArrow);}
		void Activate();
};

class EW_DecrButton : public EW_Button
{
	public:
		EW_DecrButton(){SetDisplayFunction(ewDfDownArrow);}
		void Activate();
};

class EW_NumberEntry : public EW_Widget
{
	private:
		EW_GetString	string_entry;
		EW_IncrButton	incr_button;
		EW_DecrButton	decr_button;
		EW_String		label;
		float			stepsize;
		float			*number;
		long			labelwidth;

	public:
		EW_NumberEntry();
		void Draw(EW_Redraw *redraw);
		void SetUp(long x,long y,long w,long h, long field_w, long button_w);
		void SetLabel(char *a_string);
		void SetStepSize(float a_stepsize){stepsize = a_stepsize;}
		void SetNumber(float *a_float){number = a_float;}
		void SetNumberValue(float a_float){*number = a_float;}
		float *GetNumber(){return number;}
		void Increment();
		void Decrement();
};
