/******************************************************************************

Coral Tree dna/client.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <wba/wba.h>
#include <dna/tcp.h>
#include <dna/client.h>

/******************************************************************************
class DNA_Client
******************************************************************************/

/******************************************************************************
DNA_Client::DNA_Client(void)
******************************************************************************/
DNA_Client::DNA_Client(void) : timeout(0)
{
	// Provide a default implementation of TCP/IP.
	connection.SetImp((DNA_BusImp *) new DNA_TCPConnectionImp());

	// Set the source address, since it is always this process.
	register DNA_IPPoint *source = new DNA_IPPoint();
	if ((DNA_IPPoint *) GPL_BAD_ADDRESS == source)
	{
		wbaLog("DNA ERROR", "out of memory");
	}
	GPL_String host;
	if (DNA_IPPoint::GetHost(host))
	{
		source->SetHost(host);
		connection.SetSource(source);
	}
}

/******************************************************************************
GPL_Boolean DNA_Client::Initialize(void)
******************************************************************************/
GPL_Boolean DNA_Client::Initialize(void)
{
	if (!connection.Open())
	{
		return FALSE;
	}
	if (!connection.Bind())
	{
		return FALSE;
	}
	if (!connection.Connect())
	{
		return FALSE;
	}
	return TRUE;
}

/******************************************************************************
void DNA_Client::Finalize(void)
******************************************************************************/
void DNA_Client::Finalize(void)
{
	connection.Disconnect();
	connection.Close();
}

/******************************************************************************
GPL_Boolean DNA_Client::Reconnect(void)
******************************************************************************/
GPL_Boolean DNA_Client::Reconnect(void)
{
	DNA_IPPoint *s = (DNA_IPPoint *) connection.GetSource()->Clone();
	DNA_IPPoint *d = (DNA_IPPoint *) connection.GetDestination()->Clone();
	DNA_UDPBindingImp *i = (DNA_UDPBindingImp *) connection.GetBindingImp()->Clone();
	connection = DNA_Connection();
	connection.SetImp(i);
	connection.SetSource(s);
	connection.SetDestination(d);
	if (!connection.Open())
	{
		return FALSE;
	}
	if (!connection.Bind())
	{
		return FALSE;
	}
	if (!connection.Connect())
	{
		return FALSE;
	}
	return TRUE;
}

/******************************************************************************
void DNA_Client::Process(void)
******************************************************************************/
void DNA_Client::Process(void)
{
	DNA_Message message;
	if (message.Receive(&connection, timeout))
	{
		do
		{
			const GPL_String image = message.ToString();
			wbaLog("DNA LOG", "RCV >%s<", image.GetData());
			ProcessMessage(message);
		}
		while (message.Receive(&connection, 0));
	}
	else
	{
		// Reap any expired handlers.
		register unsigned long i = 0;
		while (handlers.GetCount() > i)
		{
			if (handlers[i]->GetExpired())
			{
				handlers.Remove(i);
			}
			else
			{
				++i;
			}
		}
	}
}

/******************************************************************************

******************************************************************************/
GPL_Boolean DNA_Client::Receive(DNA_Message & message)
{
	if(message.Receive(&connection, timeout))
	{
		return TRUE;
	}
	return FALSE;
}

/******************************************************************************
void DNA_Client::ProcessMessage(DNA_Message & message)
******************************************************************************/
void DNA_Client::ProcessMessage(DNA_Message & message)
{
	GPL_Boolean handledMessage = FALSE;
	GPL_Boolean consumedMessage = FALSE;
	const long originalPosition = message.Position(0);
	register unsigned long i = 0;
	while (handlers.GetCount() > i)
	{
		// Give the handler a chance to handle the message.
		consumedMessage = FALSE;
		handledMessage = handlers[i]->HandleMessage(message, consumedMessage);

		// Reap an expired handler.
		if (handlers[i]->GetExpired())
		{
			handlers.Remove(i);
		}
		else
		{
			++i;
		}

		// Successful handlers which mark the message
		// as consumed terminate the loop.
		if (consumedMessage)
		{
			if (handledMessage)
			{
				break;
			}
		}

		// Otherwise, restore the message to its original state.
		else
		{
			message.Position(originalPosition - message.Position(0));
		}
	}
}

/******************************************************************************
GPL_Boolean DNA_Client::Request(DNA_Message & message)
******************************************************************************/

/******************************************************************************
GPL_Boolean DNA_Client::Request(DNA_Message & message)
******************************************************************************/
GPL_Boolean DNA_Client::Request(DNA_Message & message)
{
	const GPL_String image = message.ToString();
	wbaLog("DNA LOG", "SND >%s<", image.GetData());
	return message.Send(&connection);
}

/******************************************************************************
GPL_Boolean DNA_Client::Reply(DNA_Message & message)
******************************************************************************/
GPL_Boolean DNA_Client::Reply(DNA_Message & message)
{
	const GPL_String image = message.ToString();
	wbaLog("DNA LOG", "SND >%s<", image.GetData());
	return message.Send(&connection);
}

/******************************************************************************
unsigned long DNA_Client::GetNumberHandlers(void) const
******************************************************************************/
unsigned long DNA_Client::GetNumberHandlers(void) const
{
	return handlers.GetCount();
}

/******************************************************************************
void DNA_Client::AppendHandler(const GPL_Pointer<DNA_MessageHandler> & handler)
******************************************************************************/
void DNA_Client::AppendHandler(const GPL_Pointer<DNA_MessageHandler> & handler)
{
	if (!handler.IsNull())
	{
		handlers.Append(handler);
	}
}

/******************************************************************************
void DNA_Client::InsertHandler(const GPL_Pointer<DNA_MessageHandler> & handler,
								const unsigned long at)
******************************************************************************/
void DNA_Client::InsertHandler(const GPL_Pointer<DNA_MessageHandler> & handler,
								const unsigned long at)
{
	if (!handler.IsNull())
	{
		handlers.Insert(handler, at);
	}
}

