
/******************************************************************************

Coral Tree dna/message.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/

#include <dna/dna.h>

//TODO: remove the dependency on GPL_Message's protected (should be private)
//      data.  Currently, the InsertLong on the Send is copying the data
//      anyway, so there is no real performance gain by having such access.

/******************************************************************************
DNA_Message::DNA_Message(void)
******************************************************************************/
DNA_Message::DNA_Message(void)
{
}

/******************************************************************************
DNA_Message::DNA_Message(const DNA_Message & msg)
******************************************************************************/
DNA_Message::DNA_Message(const DNA_Message & msg)
{
	operator=(msg);
}

/******************************************************************************
DNA_Message::~DNA_Message(void)
******************************************************************************/
DNA_Message::~DNA_Message(void)
{
}

/******************************************************************************
DNA_Message & DNA_Message::operator=(const DNA_Message & msg)
******************************************************************************/
DNA_Message & DNA_Message::operator=(const DNA_Message & msg)
{
	GPL_Message::operator=(msg);
	return *this;
}

/******************************************************************************
GPL_Boolean DNA_Message::Send(DNA_Bus *bus)
******************************************************************************/
GPL_Boolean DNA_Message::Send(DNA_Bus *bus)
{
	// Temporarily insert the over-all message length.
	InsertLong(length,0);

	// Write the message to the bus.
	unsigned long n = 0;
	if (!bus->Write(buffer, length, n) || n != (unsigned long) length)
	{
		return FALSE;
	}

	// Remove the over-all message length.
	DeleteRange(0, 3);

	return TRUE;
}

/******************************************************************************
GPL_Boolean DNA_Message::Receive(DNA_Bus *bus, const long timeout)
******************************************************************************/
GPL_Boolean DNA_Message::Receive(DNA_Bus *bus, const long timeout)
{
	// Clear any previous contents.
	Clear();

	// Read the message length from the bus.
	unsigned long n = 0;
	GPL_Array<char> temp(sizeof(unsigned long));
	if (!bus->Read(temp.GetData(), temp.GetSize(), timeout, n))
	{
		return FALSE;
	}
	if (temp.GetSize() != n)
	{
		wbaLog("DNA WARNING", "Received corrupt message");
		return FALSE;
	}

	// Convert the message length.
	// NOTE: Not MT-Safe!
	// static GPL_Message converter;
	// Now it is
	GPL_Message converter;
	converter.SetBuffer(temp.GetData(), temp.GetSize(), FALSE);
	length = converter.ReadUnsignedLong();
	CheckSize(length);

	// If this is a read-only or copied buffer of
	// insufficient length, make a separate buffer into
	// which the bytes can be read.  This keeps the
	// network clean and avoids polluting the heap.
	register char *buf = buffer;
	if (GPL_MESSAGE_READONLY & flags ||
		((GPL_MESSAGE_INVALID & flags) && (GPL_MESSAGE_COPY & flags)))
	{
		temp.Resize(length);
		buf = temp.GetData();
	}

	if (!bus->Read(buf, length, timeout, n))
	{
		return FALSE;
	}
	if ((unsigned long) length != n)
	{
		wbaLog("DNA WARNING", "Received corrupt message");
		return FALSE;
	}

	// If this is a read-only message...
	if (GPL_MESSAGE_READONLY & flags)
	{
		return FALSE;
	}
	
	// Copy the appropriate number of bytes out of the
	// temporary buffer into the message.  The caller
	// is responsible for checking GPL_MESSAGE_INVALID
	// to determine if the incoming message exceeded
	// the ability of the copied buffer to hold.
	if (buffer != buf)
	{
		osdMemcpy(buffer, buf, length);
	}

	return TRUE;
}

