/******************************************************************************

Coral Tree dna/rmfbus.cc
Copyright (C) 1998 Infinite Monkeys Incorporated

This program is free software; you can redistribute it and/or modify
it under the terms of the Artistic License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
Artistic License for more details.

You should have received a copy of the Artistic License along with
this program (see meta/ART_LICE); if not, write to coral@imonk.com.

******************************************************************************/



#include <dna/dna.h>


#define DNA_RMF_VERBOSE	TRUE


#ifdef DNA_RMF


/******************************************************************************
		DNA_RMF_Group_BusImp::DNA_RMF_Group_BusImp(void)

******************************************************************************/
DNA_RMF_Group_BusImp::DNA_RMF_Group_BusImp(void)
	{
	sender=new RAMP();
	receiver=new RMF_UniversalReceiver();
	}



/******************************************************************************
		DNA_RMF_Group_BusImp::~DNA_RMF_Group_BusImp(void)

******************************************************************************/
DNA_RMF_Group_BusImp::~DNA_RMF_Group_BusImp(void)
	{
	DNA_Error error;
	if(opened) Close(error);
	delete sender;
	delete receiver;
	}



/******************************************************************************
void	DNA_RMF_Group_BusImp::Reset(void)

******************************************************************************/
void DNA_RMF_Group_BusImp::Reset(void)
	{
	DNA_Error error;

	Close(error);
	}



/******************************************************************************
long	DNA_RMF_Group_BusImp::Open(DNA_Error & error)

	returns FALSE on error

******************************************************************************/
long DNA_RMF_Group_BusImp::Open(DNA_Error & error)
{
	DNA_RMF_Config config;
	return Open(error, config);
}

long DNA_RMF_Group_BusImp::Open(DNA_Error & error,  DNA_RMF_Config &config)
	{
	RMF_Address group;

	group.SetAddress(ntohl(GetGroupPoint().GetAddress()));
	group.SetPort(GetGroupPoint().GetPort());

	if(opened)
		{
		Close(error);
		}

	if( sender->SetOption(RAMP_RELIABILITY_MODE, config.reliabilityMode)!= RMF_OK)
		{
#if DNA_RMF_VERBOSE
		WBA_APP->Printf("DNA_RMF_Group_BusImp::Open() RAMP::SetOption() failed");
#endif
		error.Set(DNA_ERROR_UNKNOWN);
		return FALSE;
		}

	if( sender->SetOption(RAMP_WINDOW_SIZE, config.rampWindowSize)!= RMF_OK)
		{
#if DNA_RMF_VERBOSE
		WBA_APP->Printf("DNA_RMF_Group_BusImp::Open() RAMP::SetOption() failed");
#endif
		error.Set(DNA_ERROR_UNKNOWN);
		return FALSE;
		}

	if( sender->Open(group,RMF_SCOPE_SUBNET) != RMF_OK)
		{
#if DNA_RMF_VERBOSE
		WBA_APP->Printf("DNA_RMF_Group_BusImp::Open() RAMP::Open() failed");
#endif
		error.Set(DNA_ERROR_UNKNOWN);
		return FALSE;
		}

	if(receiver->SetOption(RMF_RELIABLE_RCV_MODE,RMF_RCV_ORDERED_DATAGRAM)!= RMF_OK)
		{
#if DNA_RMF_VERBOSE
		WBA_APP->Printf("DNA_RMF_Group_BusImp::Open() RMF_UniversalReceiver::SetOption() failed");
#endif
		error.Set(DNA_ERROR_UNKNOWN);
		return FALSE;
		}

	if( receiver->Join(group,RMF_SCOPE_SUBNET) != RMF_OK)
		{
#if DNA_RMF_VERBOSE
		WBA_APP->Printf("DNA_RMF_Group_BusImp::Open() RMF_UniversalReceiver::Join() failed");
#endif
		error.Set(DNA_ERROR_UNKNOWN);
		return FALSE;
		}

	opened=TRUE;

	return TRUE;
	}



/******************************************************************************
void	DNA_RMF_Group_BusImp::Close(DNA_Error & error)

	returns FALSE on error

******************************************************************************/
void DNA_RMF_Group_BusImp::Close(DNA_Error & error)
	{
	if(opened)
		{
		opened=FALSE;

		receiver->Leave();
		sender->Close();
		}
	}



/******************************************************************************
long	DNA_RMF_Group_BusImp::Compare(const DNA_BusImp *imp) const

******************************************************************************/
long DNA_RMF_Group_BusImp::Compare(const DNA_BusImp *imp) const
	{
	return 0;
	}



/******************************************************************************
GPL_Boolean DNA_RMF_Group_BusImp::Read(void *buffer,const unsigned long size,
									unsigned long & length,DNA_Error & error)

******************************************************************************/
GPL_Boolean DNA_RMF_Group_BusImp::Read(void *buffer,const unsigned long size,unsigned long & length,
																					DNA_Error & error)
	{
	return Read(buffer,size,-1,length,error);
	}



/******************************************************************************
GPL_Boolean DNA_RMF_Group_BusImp::Read(void *buffer,const unsigned long size,
				const long timeout,unsigned long & length,DNA_Error & error)

******************************************************************************/
GPL_Boolean DNA_RMF_Group_BusImp::Read(void *buffer,const unsigned long size,const long timeout,
															unsigned long & length,DNA_Error & error)
	{
	long result=TRUE;

	RMF_Buffer *rmfbuffer=new RMF_Buffer();

	rmfbuffer->buffer=(UINT8 *)buffer;
	rmfbuffer->offset=0;
	rmfbuffer->length=0;
	rmfbuffer->maxLength=size;

	if(receiver->SetOption(RMF_RCV_NODELAY,timeout<0? RMF_FALSE: RMF_TRUE) != RMF_OK)
		{
#if DNA_RMF_VERBOSE
		WBA_APP->Printf("DNA_RMF_Group_BusImp::Read() RMF_UniversalReceiver::SetOption() failed");
#endif
		error.Set(DNA_ERROR_UNKNOWN);
		result=FALSE;
		}

	if(receiver->Receive(*rmfbuffer) != RMF_OK)
		{
#if DNA_RMF_VERBOSE
		WBA_APP->Printf("DNA_RMF_Group_BusImp::Read() RMF_UniversalReceiver::Receive() failed");
#endif
		error.Set(DNA_ERROR_UNKNOWN);
		result=FALSE;
		}

	length=rmfbuffer->length;

	SetLastSendersPoint(0,0);
	if(result && length>0)
		{
		UINT32 ownerid=receiver->ReceiveDataOwner();

		RMF_Address rmfaddress;
		if(receiver->GetAddressFromID(ownerid,rmfaddress) != RMF_OK)
			{
#if DNA_RMF_VERBOSE
	WBA_APP->Printf("DNA_RMF_Group_BusImp::Read() RMF_UniversalReceiver::GetAddressFromID() failed");
#endif
			error.Set(DNA_ERROR_UNKNOWN);
			result=FALSE;
			}
		else
			SetLastSendersPoint(rmfaddress.GetAddress(),rmfaddress.GetPort());
		}

	// in case RMF_Buffer destructor tries to delete buffer
	rmfbuffer->buffer=NULL;

	delete rmfbuffer;

	return result;
	}



/******************************************************************************
GPL_Boolean	DNA_RMF_Group_BusImp::Write(void *buffer,const unsigned long size,
									unsigned long & length,DNA_Error & error)

******************************************************************************/
GPL_Boolean DNA_RMF_Group_BusImp::Write(void *buffer,const unsigned long size,unsigned long & length, DNA_Error & error)
	{
	long result=TRUE;

	//RMF_Buffer *rmfbuffer=new RMF_Buffer();
	RMF_Buffer stackBuffer;
	RMF_Buffer *rmfbuffer = &stackBuffer;

	rmfbuffer->buffer=(UINT8 *)buffer;
	rmfbuffer->offset=0;
	rmfbuffer->length=size;
	rmfbuffer->maxLength=size;

	if(sender->Send(*rmfbuffer) != RMF_OK)
		{
#if DNA_RMF_VERBOSE
		WBA_APP->Printf("DNA_RMF_Group_BusImp::Write() RAMP::Send() failed");
#endif
		error.Set(DNA_ERROR_UNKNOWN);
		result=FALSE;
		}

	// in case RMF_Buffer destructor tries to delete buffer
	rmfbuffer->buffer=NULL;

	//delete rmfbuffer;

	length= (result)? size: 0;

	return result;
	}

GPL_OBJECT_IMPLEMENTATION(DNA_RMF_Group_BusImp, 0x40070009, "DNA_RMF_Group_BusImp");

#endif /* DNA_RMF */
